@@ -4,10 +4,10 @@
  *    Remove cr ('\x0d') characters from a file.
  *
  *  The dos2unix package is distributed under FreeBSD style license.
- *  See also http://www.freebsd.org/copyright/freebsd-license.html
+ *  See also https://www.freebsd.org/copyright/freebsd-license.html
  *  --------
  *
- *  Copyright (C) 2009-2016 Erwin Waterlander
+ *  Copyright (C) 2009-2025 Erwin Waterlander
  *  Copyright (C) 1998 Christian Wurll
  *  Copyright (C) 1998 Bernd Johannes Wuebben
  *  Copyright (C) 1994-1995 Benjamin Lin.
@@ -81,7 +81,7 @@ Copyright (C) 2009-%d Erwin Waterlander\n\
 Copyright (C) 1998      Christian Wurll (Version 3.1)\n\
 Copyright (C) 1998      Bernd Johannes Wuebben (Version 3.0)\n\
 Copyright (C) 1994-1995 Benjamin Lin\n\
-All rights reserved.\n\n"),2016);
+All rights reserved.\n\n"),2025);
   PrintBSDLicense();
 }
 
@@ -180,6 +180,7 @@ int StripDelimiter(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, int CurChar, unsign
 int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progname)
 {
     int RetVal = 0;
+    wint_t PrevChar = WEOF;
     wint_t TempChar;
     wint_t TempNextChar;
     unsigned int line_nr = 1;
@@ -207,7 +208,7 @@ int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *prog
             RetVal = -1;
             ipFlag->status |= BINARY_FILE ;
             if (ipFlag->verbose) {
-              if ((ipFlag->stdio_mode) && (!ipFlag->error)) ipFlag->error = 1;
+              ipFlag->error = 1;
               D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
               D2U_UTF8_FPRINTF(stderr, _("Binary symbol 0x00%02X found at line %u\n"),TempChar, line_nr);
             }
@@ -227,6 +228,18 @@ int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *prog
               break;
             }
           }
+          PrevChar = TempChar;
+        }
+        if (TempChar == WEOF && ipFlag->add_eol && PrevChar != WEOF && PrevChar != 0x0a) {
+          /* Add missing line break at the last line. */
+            if (ipFlag->verbose > 1) {
+              D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+              D2U_UTF8_FPRINTF(stderr, _("Added line break to last line.\n"));
+            }
+            if (d2u_putwc(0x0a, ipOutF, ipFlag, progname) == WEOF) {
+              RetVal = -1;
+              d2u_putwc_error(ipFlag,progname);
+            }
         }
         if ((TempChar == WEOF) && ferror(ipInF)) {
           RetVal = -1;
@@ -244,22 +257,21 @@ int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *prog
             RetVal = -1;
             ipFlag->status |= BINARY_FILE ;
             if (ipFlag->verbose) {
-              if ((ipFlag->stdio_mode) && (!ipFlag->error)) ipFlag->error = 1;
+              ipFlag->error = 1;
               D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
               D2U_UTF8_FPRINTF(stderr, _("Binary symbol 0x00%02X found at line %u\n"), TempChar, line_nr);
             }
             break;
           }
           if ((TempChar != 0x0d)) {
-              if (TempChar == 0x0a) /* Count all DOS and Unix line breaks */
-                ++line_nr;
-              if(d2u_putwc(TempChar, ipOutF, ipFlag, progname) == WEOF) {
-                RetVal = -1;
-                d2u_putwc_error(ipFlag,progname);
-                break;
-              }
+            if (TempChar == 0x0a) /* Count all DOS and Unix line breaks */
+              ++line_nr;
+            if(d2u_putwc(TempChar, ipOutF, ipFlag, progname) == WEOF) {
+              RetVal = -1;
+              d2u_putwc_error(ipFlag,progname);
+              break;
             }
-          else{
+          } else{
             /* TempChar is a CR */
             if ( (TempNextChar = d2u_getwc(ipInF, ipFlag->bomtype)) != WEOF) {
               if (d2u_ungetwc( TempNextChar, ipInF, ipFlag->bomtype) == WEOF) {  /* put back peek char */
@@ -274,14 +286,15 @@ int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *prog
                   RetVal = -1;
                   break;
                 }
+                PrevChar = TempChar;
                 continue;
               }
             }
             if (d2u_putwc(0x0a, ipOutF, ipFlag, progname) == WEOF) { /* MAC line end (CR). Put LF */
-                RetVal = -1;
-                d2u_putwc_error(ipFlag,progname);
-                break;
-              }
+              RetVal = -1;
+              d2u_putwc_error(ipFlag,progname);
+              break;
+            }
             converted++;
             line_nr++; /* Count all Mac line breaks */
             if (ipFlag->NewLine) {  /* add additional LF? */
@@ -292,6 +305,18 @@ int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *prog
               }
             }
           }
+          PrevChar = TempChar;
+        }
+        if (TempChar == WEOF && ipFlag->add_eol && PrevChar != WEOF && !(PrevChar == 0x0a || PrevChar == 0x0d)) {
+          /* Add missing line break at the last line. */
+            if (ipFlag->verbose > 1) {
+              D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+              D2U_UTF8_FPRINTF(stderr, _("Added line break to last line.\n"));
+            }
+            if (d2u_putwc(0x0a, ipOutF, ipFlag, progname) == WEOF) {
+              RetVal = -1;
+              d2u_putwc_error(ipFlag,progname);
+            }
         }
         if ((TempChar == WEOF) && ferror(ipInF)) {
           RetVal = -1;
@@ -323,6 +348,7 @@ int ConvertDosToUnixW(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *prog
 int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progname)
 {
     int RetVal = 0;
+    int PrevChar = EOF;
     int TempChar;
     int TempNextChar;
     int *ConvTable;
@@ -390,7 +416,7 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
             RetVal = -1;
             ipFlag->status |= BINARY_FILE ;
             if (ipFlag->verbose) {
-              if ((ipFlag->stdio_mode) && (!ipFlag->error)) ipFlag->error = 1;
+              ipFlag->error = 1;
               D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
               D2U_UTF8_FPRINTF(stderr, _("Binary symbol 0x%02X found at line %u\n"),TempChar, line_nr);
             }
@@ -410,6 +436,18 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
               break;
             }
           }
+          PrevChar = TempChar;
+        }
+        if (TempChar == EOF && ipFlag->add_eol && PrevChar != EOF && PrevChar != '\x0a') {
+          /* Add missing line break at the last line. */
+            if (ipFlag->verbose > 1) {
+              D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+              D2U_UTF8_FPRINTF(stderr, _("Added line break to last line.\n"));
+            }
+           if (fputc('\x0a', ipOutF) == EOF) {
+              RetVal = -1;
+              d2u_putc_error(ipFlag,progname);
+            }
         }
         if ((TempChar == EOF) && ferror(ipInF)) {
           RetVal = -1;
@@ -427,22 +465,21 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
             RetVal = -1;
             ipFlag->status |= BINARY_FILE ;
             if (ipFlag->verbose) {
-              if ((ipFlag->stdio_mode) && (!ipFlag->error)) ipFlag->error = 1;
+              ipFlag->error = 1;
               D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
               D2U_UTF8_FPRINTF(stderr, _("Binary symbol 0x%02X found at line %u\n"),TempChar, line_nr);
             }
             break;
           }
           if ((TempChar != '\x0d')) {
-              if (TempChar == '\x0a') /* Count all DOS and Unix line breaks */
-                ++line_nr;
-              if(fputc(ConvTable[TempChar], ipOutF) == EOF) {
-                RetVal = -1;
-                d2u_putc_error(ipFlag,progname);
-                break;
-              }
+            if (TempChar == '\x0a') /* Count all DOS and Unix line breaks */
+              ++line_nr;
+            if(fputc(ConvTable[TempChar], ipOutF) == EOF) {
+              RetVal = -1;
+              d2u_putc_error(ipFlag,progname);
+              break;
             }
-          else{
+          } else{
             /* TempChar is a CR */
             if ( (TempNextChar = fgetc(ipInF)) != EOF) {
               if (ungetc( TempNextChar, ipInF ) == EOF) {  /* put back peek char */
@@ -457,14 +494,15 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
                   d2u_putc_error(ipFlag,progname);
                   break;
                 }
+                PrevChar = TempChar;
                 continue;
               }
             }
             if (fputc('\x0a', ipOutF) == EOF) { /* MAC line end (CR). Put LF */
-                RetVal = -1;
-                d2u_putc_error(ipFlag,progname);
-                break;
-              }
+              RetVal = -1;
+              d2u_putc_error(ipFlag,progname);
+              break;
+            }
             converted++;
             line_nr++; /* Count all Mac line breaks */
             if (ipFlag->NewLine) {  /* add additional LF? */
@@ -475,6 +513,18 @@ int ConvertDosToUnix(FILE* ipInF, FILE* ipOutF, CFlag *ipFlag, const char *progn
               }
             }
           }
+          PrevChar = TempChar;
+        }
+        if (TempChar == EOF && ipFlag->add_eol && PrevChar != EOF && !(PrevChar == '\x0a' || PrevChar == '\x0d')) {
+          /* Add missing line break at the last line. */
+            if (ipFlag->verbose > 1) {
+              D2U_UTF8_FPRINTF(stderr, "%s: ", progname);
+              D2U_UTF8_FPRINTF(stderr, _("Added line break to last line.\n"));
+            }
+            if (fputc('\x0a', ipOutF) == EOF) {
+              RetVal = -1;
+              d2u_putc_error(ipFlag,progname);
+            }
         }
         if ((TempChar == EOF) && ferror(ipInF)) {
           RetVal = -1;
@@ -504,11 +554,16 @@ int main (int argc, char *argv[])
   CFlag *pFlag;
   char *ptr;
   char localedir[1024];
+  int ret;
 # ifdef __MINGW64__
   int _dowildcard = -1; /* enable wildcard expansion for Win64 */
 # endif
   int  argc_new;
   char **argv_new;
+#ifdef D2U_UNIFILE
+  wchar_t **wargv;
+  char ***argv_glob;
+# endif
 
   progname[8] = '\0';
   strcpy(progname,"dos2unix");
@@ -516,14 +571,14 @@ int main (int argc, char *argv[])
 #ifdef ENABLE_NLS
    ptr = getenv("DOS2UNIX_LOCALEDIR");
    if (ptr == NULL)
-      strcpy(localedir,LOCALEDIR);
+      d2u_strncpy(localedir,LOCALEDIR,sizeof(localedir));
    else {
       if (strlen(ptr) < sizeof(localedir))
-         strcpy(localedir,ptr);
+         d2u_strncpy(localedir,ptr,sizeof(localedir));
       else {
          D2U_UTF8_FPRINTF(stderr,"%s: ",progname);
          D2U_ANSI_FPRINTF(stderr, "%s", _("error: Value of environment variable DOS2UNIX_LOCALEDIR is too long.\n"));
-         strcpy(localedir,LOCALEDIR);
+         d2u_strncpy(localedir,LOCALEDIR,sizeof(localedir));
       }
    }
 #endif
@@ -569,8 +624,6 @@ int main (int argc, char *argv[])
 
 #ifdef D2U_UNIFILE
   /* Get arguments in wide Unicode format in the Windows Command Prompt */
-  wchar_t **wargv;
-  char ***argv_glob;
 
   /* This does not support wildcard expansion (globbing) */
   wargv = CommandLineToArgvW(GetCommandLineW(), &argc);
@@ -579,6 +632,7 @@ int main (int argc, char *argv[])
   if (argv_glob == NULL) {
     D2U_UTF8_FPRINTF(stderr, "%s:", progname);
     D2U_ANSI_FPRINTF(stderr, " %s\n", strerror(errno));
+    free(pFlag);
     return errno;
   }
   /* Glob the arguments and convert them to UTF-8 */
@@ -590,9 +644,10 @@ int main (int argc, char *argv[])
 #endif
 
 #ifdef D2U_UNICODE
-  return parse_options(argc_new, argv_new, pFlag, localedir, progname, PrintLicense, ConvertDosToUnix, ConvertDosToUnixW);
+  ret = parse_options(argc_new, argv_new, pFlag, localedir, progname, PrintLicense, ConvertDosToUnix, ConvertDosToUnixW);
 #else
-  return parse_options(argc_new, argv_new, pFlag, localedir, progname, PrintLicense, ConvertDosToUnix);
+  ret = parse_options(argc_new, argv_new, pFlag, localedir, progname, PrintLicense, ConvertDosToUnix);
 #endif
+  free(pFlag);
+  return ret;
 }
-
